// Boost pinger (c) 2011 by Guilherme Maciel Ferreira / Intra2net AG
// Based upon work copyright (c) 2003-2010 Christopher M. Kohlhoff (ping.cpp)
//
// Distributed under the Boost Software License, Version 1.0.
//    (See accompanying file LICENSE_1_0.txt or copy at
//          http://www.boost.org/LICENSE_1_0.txt)
#include "icmp/icmppinger.h"

#include <errno.h>

#include <ostream>

#include <boost/bind.hpp>
#include <boost/date_time/posix_time/posix_time.hpp>
#include <boost/date_time/posix_time/posix_time_types.hpp>
#include <boost/uuid/uuid.hpp>
#include <boost/uuid/uuid_generators.hpp>

#include <logfunc.hpp>

#include "boost_assert_handler.h"
#include "icmp/icmppacketfactory.h"

using namespace std;
using boost::asio::const_buffers_1;
using boost::asio::io_service;
using boost::asio::ip::address;
using boost::asio::ip::icmp;
using boost::function;
using boost::posix_time::microsec_clock;
using boost::posix_time::seconds;
using boost::shared_ptr;
using I2n::Logger::GlobalLogger;

//-----------------------------------------------------------------------------
// IcmpPinger
//-----------------------------------------------------------------------------

/**
 * @brief factory function for IcmpPingers, ensures that set_myself is set
 *
 * @returns a shared pointer to a Pinger
 */
PingerItem IcmpPinger::create(
        const IoServiceItem io_serv,
        const icmp::socket::protocol_type &protocol,
        const string &source_network_interface,
        const int echo_reply_timeout_in_sec )
{
    // get distributor
    IcmpPacketDistributorItem distributor = IcmpPacketDistributor::get_distributor(
            icmp::v4(), source_network_interface, io_serv);

    // create pinger
    IcmpPinger *ptr = new IcmpPinger(io_serv, protocol, echo_reply_timeout_in_sec);
    IcmpPingerItem shared_ptr_(ptr);
    Pinger::WeakPtr weak_ptr( shared_ptr_ );

    // keep weak pointer to self
    //shared_ptr_->set_myself( weak_ptr ); //Error: Pinger::set_myself is protected
    ptr->set_myself( weak_ptr );

    // register in distributor
    distributor->register_pinger(shared_ptr_);

    // done, return shared ptr
    return shared_ptr_;
}

/**
 * @brief Parameterized constructor.
 *
 * @param io_serv The one @c io_service object that controls async processing
 * @param protocol The network layer protocol to use.
 * @param source_network_interface The network interface name from where to
 * send the packets.
 * @param echo_reply_timeout_in_sec The amount of time to wait for a reply.
 */
IcmpPinger::IcmpPinger(
        const IoServiceItem io_serv,
        const icmp::socket::protocol_type &protocol,
        const int echo_reply_timeout_in_sec
) :
    DestinationEndpoint(),
    Protocol( protocol ),
    IcmpPacketReceiveTimer( *io_serv ),
    Identifier( 0 ),
    SequenceNumber( 0 ),
    TimeSent( microsec_clock::universal_time() ),
    ReplyReceived( false ),
    EchoReplyTimeoutInSec( echo_reply_timeout_in_sec ),
    PingerStatus( PingStatus_NotSent ),
    PingDoneCallback()
{
    // Create "unique" identifier
    boost::uuids::random_generator random_gen;
    boost::uuids::uuid random_tag = random_gen();

    BOOST_ASSERT( sizeof(Identifier) <= random_tag.size() );
    memcpy( &Identifier, random_tag.data, sizeof(Identifier) );
}

/**
 * @brief Destructor.
 */
IcmpPinger::~IcmpPinger()
{
}

/**
 * @brief Ping a destination address from an available local source.
 *
 * @param destination_ip The address of the host to ping.
 * @param destination_port The port at the destination host to ping.
 * @param done_handler Done handler will be called on successful ping or timeout.
 *
 * @return void.
 */
void IcmpPinger::ping(
        const string &destination_ip,
        const uint16_t /*destination_port*/, // the ICMP protocol does not use ports
        function<void(bool)> ping_done_callback
)
{
    BOOST_ASSERT( !destination_ip.empty() );

    PingDoneCallback = ping_done_callback;

    // Prepare ping
    set_ping_status( PingStatus_NotSent );

    set_destination_endpoint( destination_ip );

    start_send();
}

void IcmpPinger::stop_pinging()
{
    IcmpPingerItem icmp_item = boost::const_pointer_cast<IcmpPinger>( get_myself().lock() );
    PacketDistributor->unregister_pinger( icmp_item );
}


void IcmpPinger::set_destination_endpoint( const string &destination_ip )
{
    BOOST_ASSERT( !destination_ip.empty() );

    uint16_t port = 0;
    address destination_address = address::from_string( destination_ip );
    DestinationEndpoint = icmp::endpoint( destination_address, port );
}

bool IcmpPinger::start_send()
{
    ++SequenceNumber;

    IcmpPacketItem icmp_packet_echo_request = IcmpPacketFactory::create_icmp_packet_echo_request(
            Protocol, Identifier, SequenceNumber );

    BOOST_ASSERT( PingerStatus == PingStatus_NotSent );
    return send_echo_request( icmp_packet_echo_request );
}

bool IcmpPinger::send_echo_request( const IcmpPacketItem icmp_packet )
{
    boost::asio::streambuf request_buffer;
    ostream os( &request_buffer );
    if ( !icmp_packet->write( os ) )
    {
        GlobalLogger.error()
           << DestinationEndpoint.address().to_string()
           << ": fail writing ping data." << endl;
    }

    TimeSent = microsec_clock::universal_time();

    string dest_address_string = DestinationEndpoint.address().to_string();
    BOOST_ASSERT( !dest_address_string.empty() );

    // Send the request
    size_t bytes_sent = 0;
    try
    {
        GlobalLogger.info()
                    << DestinationEndpoint.address().to_string()
                    << ": sending ping" << endl;
        const_buffers_1 data = request_buffer.data();
        // Block until send the data
        bytes_sent = PacketDistributor->get_socket()->send_to( data, DestinationEndpoint );
        if ( bytes_sent != buffer_size( data ) )
        {
            GlobalLogger.error()
                   << DestinationEndpoint.address().to_string()
                   << ": fail sending ping data." << endl;
        }
    }
    catch ( const exception &ex )
    {
        GlobalLogger.error()
                   << DestinationEndpoint.address().to_string()
                   << ": fail sending ping data. " << ex.what() << endl;
    }

    schedule_timeout_echo_reply();

    return (bytes_sent > 0);
}

void IcmpPinger::schedule_timeout_echo_reply()
{
    // Wait up to N seconds for a reply.
    ReplyReceived = false;
    (void) IcmpPacketReceiveTimer.expires_at(
            TimeSent + seconds( EchoReplyTimeoutInSec )
    );
    IcmpPacketReceiveTimer.async_wait(
            boost::bind( &IcmpPinger::handle_timeout, this, boost::asio::placeholders::error )
    );
}

/**
 * @brief Gets called when the ping is finished: Either on timeout or on ping reply
 *
 * @return void
 **/
void IcmpPinger::handle_timeout(const boost::system::error_code& error)
{
    if (error)
    {
        if ( error ==  boost::asio::error::operation_aborted )
        {
            if (! ReplyReceived)
                GlobalLogger.notice() << "Timer waiting for ICMP echo reply was cancelled!" << endl;
            // otherwise probably called by IcmpPacketReceiveTimer.cancel in handle_receive_icmp_packet!
        }
        else
            GlobalLogger.notice() << "Error " << error << " waiting for ICMP echo reply!" << endl;

        // Still continue with rest of function, so PingStatus is updated and Callback executed
        //   when timer was cancelled
    }

    // Check ReplyReceived as the timer handler
    // is also called by Timer.cancel();
    if ( !ReplyReceived )
    {
        GlobalLogger.info()
                   << DestinationEndpoint.address().to_string()
                   << ": Request timed out" << endl;

        set_ping_status( PingStatus_FailureTimeout );
    }

    // Call ping-done handler
    bool ping_success = (PingerStatus == PingStatus_SuccessReply);
    PingDoneCallback( ping_success );
}


/**
 * @brief Receive ICMP packets
 * @param bytes_transferred Number of bytes transferred.
 * @return true if packet matches a request from this pinger, false otherwise
 **/
bool IcmpPinger::handle_receive_icmp_packet(const IcmpPacketItem icmp_packet,
                                            const size_t bytes_transferred )
{
    if ( ReplyReceived )
        // continue, might be an old packet
        // or return false right away, do not want packet anyway...
        return false;

    // We can receive all ICMP packets received by the host, so we need to
    // filter out only the echo replies that match our identifier,
    // expected sequence number, and destination host address (receive just
    // the ICMP packets from the host we had ping).

    bool does_match = false;
    if ( icmp_packet->match_echo_reply(
                            Identifier, SequenceNumber,
                            DestinationEndpoint.address() ) )
    {
        GlobalLogger.info()
           << DestinationEndpoint.address().to_string()
           << ": Received reply" << endl;

        ReplyReceived = true;
        does_match = true;

        icmp_packet->print_echo_reply( bytes_transferred, TimeSent );

        set_ping_status( PingStatus_SuccessReply );

        IcmpPacketReceiveTimer.cancel();                                        //lint !e534
    }
    else if ( icmp_packet->match_destination_unreachable(
                                 Identifier, SequenceNumber,
                                 DestinationEndpoint.address() ) )
    {
        GlobalLogger.info()
           << DestinationEndpoint.address().to_string()
           << ": Received destination unreachable" << endl;

        ReplyReceived = true;
        does_match = true;

        icmp_packet->print_destination_unreachable();

        set_ping_status( PingStatus_FailureDestinationUnreachable );

        IcmpPacketReceiveTimer.cancel();                                        //lint !e534
    }
    // Unknown ICMP reply, start another receive till timeout
    else
    {
        GlobalLogger.info()
           << DestinationEndpoint.address().to_string()
           << ": Received packet that does not match" << endl;
    }

    return does_match;
}

void IcmpPinger::set_ping_status( PingStatus ping_status )
{
    PingerStatus = ping_status;
}
